import 'dart:async';
import 'dart:io';
import 'dart:typed_data';
import 'package:mysql_client/mysql_protocol.dart';

enum _MySQLConnectionState {
  fresh,
  waitInitialHandshake,
  initialHandshakeResponseSend,
  connectionEstablished,
  waitingCommandResponse,
  quitCommandSend,
}

/// Main class to interact with MySQL database
///
/// Use [MySQLConnection.createConnection] to create connection
class MySQLConnection {
  Socket _socket;
  bool _connected = false;
  StreamSubscription<Uint8List>? _socketSubscription;
  _MySQLConnectionState _state = _MySQLConnectionState.fresh;
  final String _username;
  final String _password;
  final String? _databaseName;
  void Function(Uint8List data)? _responseCallback;
  final List<void Function()> _onCloseCallbacks = [];
  bool _inTransaction = false;
  final bool _secure;

  MySQLConnection._({
    required Socket socket,
    required String username,
    required String password,
    bool secure = true,
    String? databaseName,
  })  : _socket = socket,
        _username = username,
        _password = password,
        _databaseName = databaseName,
        _secure = secure;

  /// Creates connection with provided options
  ///
  /// [host] host to connect to
  /// [userName] database user name
  /// [password] user password
  ///
  /// By default connection uses TLS, if you don't want to use TLS, set [secure] to false
  static Future<MySQLConnection> createConnection({
    required String host,
    required int port,
    required String userName,
    required String password,
    bool secure = true,
    String? databaseName,
  }) async {
    final socket = await Socket.connect(host, port);
    socket.setOption(SocketOption.tcpNoDelay, true);

    final client = MySQLConnection._(
      socket: socket,
      username: userName,
      password: password,
      databaseName: databaseName,
      secure: secure,
    );

    return client;
  }

  /// Returns true if this connection can be used to interact with database
  bool get connected {
    return _connected;
  }

  /// Registers callack to be executed when this connection is closed
  void onClose(void Function() callback) {
    _onCloseCallbacks.add(callback);
  }

  /// Initiate connection to database
  ///
  /// Default [timeoutMs] is 5000 milliseconds
  Future<void> connect({int timeoutMs = 5000}) async {
    if (_state != _MySQLConnectionState.fresh) {
      throw Exception("Can not connect: status is not fresh");
    }

    _state = _MySQLConnectionState.waitInitialHandshake;

    _socketSubscription = _socket.listen((data) {
      for (final chunk in _splitPackets(data)) {
        _processSocketData(chunk);
      }
    });

    _socketSubscription!.onDone(() {
      _handleSocketClose();
    });

    await Future.doWhile(() async {
      if (_state == _MySQLConnectionState.connectionEstablished) {
        return false;
      }
      await Future.delayed(Duration(milliseconds: 100));
      return true;
    }).timeout(Duration(
      milliseconds: timeoutMs,
    ));
  }

  void _handleSocketClose() {
    _connected = false;
    _socket.destroy();

    for (var element in _onCloseCallbacks) {
      element();
    }
    _onCloseCallbacks.clear();
  }

  Future<void> _processSocketData(Uint8List data) async {
    if (_state == _MySQLConnectionState.waitInitialHandshake) {
      await _processInitialHandshake(data);
      return;
    }

    if (_state == _MySQLConnectionState.initialHandshakeResponseSend) {
      // check for auth switch request
      try {
        final authSwitchPacket =
            MySQLPacket.decodeAuthSwitchRequestPacket(data);

        final payload =
            authSwitchPacket.payload as MySQLPacketAuthSwitchRequest;

        switch (payload.authPluginName) {
          case 'mysql_native_password':
            final responsePayload =
                MySQLPacketAuthSwitchResponse.createWithNativePassword(
              password: _password,
              challenge: payload.authPluginData.sublist(0, 20),
            );
            final responsePacket = MySQLPacket(
              sequenceID: authSwitchPacket.sequenceID + 1,
              payload: responsePayload,
              payloadLength: 0,
            );

            _socket.add(responsePacket.encode());
            return;
          default:
            throw Exception(
                "Unsupported auth plugin name: ${payload.authPluginName}");
        }
      } catch (e) {
        // not auth switch request packet, continue packet processing
      }

      MySQLPacket packet;

      try {
        packet = MySQLPacket.decodeGenericPacket(data);
      } catch (e) {
        // print("Skipping invalid packet: $data");
        return;
      }

      if (packet.isErrorPacket()) {
        final errorPayload = packet.payload as MySQLPacketError;
        throw Exception("MySQL error: ${errorPayload.errorMessage}");
      }

      if (packet.isOkPacket()) {
        _state = _MySQLConnectionState.connectionEstablished;
        _connected = true;
      }

      return;
    }

    if (_state == _MySQLConnectionState.waitingCommandResponse) {
      _processCommandResponse(data);
      return;
    }

    throw Exception(
      "Skipping socket data, because of connection bad state\nState: ${_state.name}\nData: $data",
    );
  }

  Iterable<Uint8List> _splitPackets(Uint8List data) sync* {
    Uint8List view = data;

    while (true) {
      final packetLength = MySQLPacket.getPacketLength(view);
      final chunk = Uint8List.sublistView(view, 0, packetLength);
      yield chunk;

      view = Uint8List.sublistView(view, packetLength);

      if (view.isEmpty) {
        break;
      }
    }
  }

  Future<void> _processInitialHandshake(Uint8List data) async {
    final packet = MySQLPacket.decodeInitialHandshake(data);

    final payload = packet.payload;

    if (payload is! MySQLPacketInitialHandshake) {
      throw Exception("Expected MySQLPacketInitialHandshake packet");
    }

    if (_secure) {
      // it secure = true, initiate ssl connection
      Future<void> initiateSSL() async {
        final responsePayload = MySQLPacketSSLRequest.createDefault(
          initialHandshakePayload: payload,
          connectWithDB: _databaseName != null,
        );

        final responsePacket = MySQLPacket(
          sequenceID: 1,
          payload: responsePayload,
          payloadLength: 0,
        );

        _socket.add(responsePacket.encode());

        _socketSubscription?.pause();

        final secureSocket = await SecureSocket.secure(
          _socket,
          onBadCertificate: (certificate) => true,
        );

        // switch socket
        _socket = secureSocket;

        _socketSubscription = secureSocket.listen((data) {
          for (final chunk in _splitPackets(data)) {
            _processSocketData(chunk);
          }
        });

        _socketSubscription!.onDone(() {
          _handleSocketClose();
        });
      }

      await initiateSSL();
    }

    final authPluginName = payload.authPluginName;

    switch (authPluginName) {
      case 'mysql_native_password':
        final responsePayload =
            MySQLPacketHandshakeResponse41.createWithNativePassword(
          username: _username,
          password: _password,
          initialHandshakePayload: payload,
        );

        responsePayload.database = _databaseName;

        final responsePacket = MySQLPacket(
          payload: responsePayload,
          sequenceID: _secure ? 2 : 1,
          payloadLength: 0,
        );

        _state = _MySQLConnectionState.initialHandshakeResponseSend;
        _socket.add(responsePacket.encode());
        break;
      case 'caching_sha2_password':
        final responsePayload =
            MySQLPacketHandshakeResponse41.createWithCachingSha2Password(
          username: _username,
          password: _password,
          initialHandshakePayload: payload,
        );

        responsePayload.database = _databaseName;

        final responsePacket = MySQLPacket(
          payload: responsePayload,
          sequenceID: _secure ? 2 : 1,
          payloadLength: 0,
        );

        _state = _MySQLConnectionState.initialHandshakeResponseSend;
        _socket.add(responsePacket.encode());
        break;
      default:
        throw Exception("Unsupported auth plugin name: $authPluginName");
    }
  }

  void _processCommandResponse(Uint8List data) {
    if (_responseCallback != null) {
      _responseCallback!(data);
    }
  }

  /// Executes given [query]
  ///
  /// [execute] can be used to make any query type (SELECT, INSERT, UPDATE)
  /// You can pass named parameters using [params]
  /// Pass [iterable] true if you want to receive rows one by one in Stream fashion
  Future<IResultSet> execute(String query,
      [Map<String, dynamic>? params, bool iterable = false]) async {
    if (!_connected) {
      throw Exception("Can not execute query: connecion closed");
    }

    // wait for ready state
    if (_state != _MySQLConnectionState.connectionEstablished) {
      await _waitForState(_MySQLConnectionState.connectionEstablished)
          .timeout(Duration(seconds: 10));
    }

    _state = _MySQLConnectionState.waitingCommandResponse;

    if (params != null && params.isNotEmpty) {
      query = _substitureParams(query, params);
    }

    final payload = MySQLPacketCommQuery(query: query);

    final packet = MySQLPacket(
      sequenceID: 0,
      payload: payload,
      payloadLength: 0,
    );

    final completer = Completer<IResultSet>();

    /**
     * 0 - initial
     * 1 - columnCount decoded
     * 2 - columnDefs parsed
     * 3 - eofParsed
     * 4 - rowsParsed
     */
    int state = 0;
    int colsCount = 0;
    List<MySQLColumnDefinitionPacket> colDefs = [];
    List<MySQLResultSetRowPacket> resultSetRows = [];

    // support for iterable result set
    IterableResultSet? iterableResultSet;
    StreamSink<ResultSetRow>? sink;

    _responseCallback = (data) {
      MySQLPacket? packet;

      switch (state) {
        case 0:
          // if packet is OK packet, there is no data
          if (MySQLPacket.detectPacketType(data) == MySQLGenericPacketType.ok) {
            final okPacket = MySQLPacket.decodeGenericPacket(data);
            _state = _MySQLConnectionState.connectionEstablished;

            completer.complete(
              EmptyResultSet(okPacket: okPacket.payload as MySQLPacketOK),
            );

            return;
          }

          packet = MySQLPacket.decodeColumnCountPacket(data);
          break;
        case 1:
          packet = MySQLPacket.decodeColumnDefPacket(data);
          break;
        case 2:
          packet = MySQLPacket.decodeGenericPacket(data);
          if (packet.isEOFPacket()) {
            state = 3;
          } else if (packet.isErrorPacket()) {
            final errMsg = (packet.payload as MySQLPacketError).errorMessage;
            throw Exception("MySQL exception: $errMsg");
          } else {
            throw Exception("Unexcpected packet type");
          }
          break;
        case 3:
          if (iterable) {
            if (iterableResultSet == null) {
              iterableResultSet = IterableResultSet._(
                columns: colDefs,
              );

              sink = iterableResultSet!._sink;
              completer.complete(iterableResultSet);
            }

            // check eof
            if (MySQLPacket.detectPacketType(data) ==
                MySQLGenericPacketType.eof) {
              state = 4;

              _state = _MySQLConnectionState.connectionEstablished;
              sink!.close();
              return;
            }

            packet = MySQLPacket.decodeResultSetRowPacket(data, colsCount);
            final values = (packet.payload as MySQLResultSetRowPacket).values;
            sink!.add(ResultSetRow._(colDefs: colDefs, values: values));
            packet = null;
            break;
          } else {
            // check eof
            if (MySQLPacket.detectPacketType(data) ==
                MySQLGenericPacketType.eof) {
              state = 4;

              final resultSetPacket = MySQLPacketResultSet(
                columnCount: BigInt.from(colsCount),
                columns: colDefs,
                rows: resultSetRows,
              );

              _state = _MySQLConnectionState.connectionEstablished;
              completer.complete(ResultSet._(resultSetPacket: resultSetPacket));
              return;
            }

            packet = MySQLPacket.decodeResultSetRowPacket(data, colsCount);
            break;
          }
      }

      if (packet != null) {
        final payload = packet.payload;

        if (payload is MySQLPacketError) {
          completer.completeError("MySQL error: " + payload.errorMessage);
          return;
        } else if (payload is MySQLPacketOK) {
          // do nothing
        } else if (payload is MySQLPacketColumnCount) {
          state = 1;
          colsCount = payload.columnCount.toInt();
          return;
        } else if (payload is MySQLColumnDefinitionPacket) {
          colDefs.add(payload);
          if (colDefs.length == colsCount) {
            state = 2;
          }
        } else if (payload is MySQLResultSetRowPacket) {
          assert(iterable == false);
          resultSetRows.add(payload);
        } else {
          completer.completeError(
            "Unexpected payload received in response to COMM_QUERY request",
          );
        }
      }
    };

    _socket.add(packet.encode());

    return completer.future;
  }

  /// Execute [callback] inside database transaction
  ///
  /// If exception is thrown inside [callback] function, transaction is rolled back
  Future<T> transactional<T>(
      FutureOr<T> Function(MySQLConnection conn) callback) async {
    // prevent double transaction
    if (_inTransaction) {
      throw Exception("Already in transaction");
    }
    _inTransaction = true;

    await execute("START TRANSACTION");

    try {
      final result = await callback(this);
      await execute("COMMIT");
      return result;
    } catch (e) {
      await execute("ROLLBACK");
      _inTransaction = false;
      rethrow;
    }
  }

  String _substitureParams(String query, Map<String, dynamic> params) {
    for (final param in params.entries) {
      String value;

      if (param.value is String) {
        value = "'" + _escapeString(param.value) + "'";
      } else if (param.value is num) {
        value = param.value.toString();
      } else if (param.value is bool) {
        value = param.value ? "TRUE" : "FALSE";
      } else {
        value = "'" + _escapeString(param.value.toString()) + "'";
      }

      query = query.replaceAll(":" + param.key, value);
    }

    return query;
  }

  /// Prepares given [query]
  ///
  /// Returns [PreparedStmt] which can be used to execute prepared statement multiple times with different parameters
  /// See [PreparedStmt.execute]
  /// You shoud call [PreparedStmt.deallocate] when you don't need prepared statement anymore to prevent memory leaks
  ///
  /// Pass [iterable] true if you want to iterable result set. See [execute] for details
  Future<PreparedStmt> prepare(String query, [bool iterable = false]) async {
    if (!_connected) {
      throw Exception("Can not prepare stmt: connecion closed");
    }

    // wait for ready state
    if (_state != _MySQLConnectionState.connectionEstablished) {
      await _waitForState(_MySQLConnectionState.connectionEstablished)
          .timeout(Duration(seconds: 10));
    }

    _state = _MySQLConnectionState.waitingCommandResponse;

    final payload = MySQLPacketCommStmtPrepare(query: query);

    final packet = MySQLPacket(
      sequenceID: 0,
      payload: payload,
      payloadLength: 0,
    );

    final completer = Completer<PreparedStmt>();

    /**
     * 0 - initial
     * 1 - first packet decoded
     * 2 - eof decoded
     */
    int state = 0;
    MySQLPacketStmtPrepareOK? preparedPacket;

    _responseCallback = (data) {
      MySQLPacket? packet;

      switch (state) {
        case 0:
          packet = MySQLPacket.decodeCommPrepareStmtResponsePacket(data);
          state = 1;
          break;
        default:
          if (MySQLPacket.detectPacketType(data) ==
              MySQLGenericPacketType.eof) {
            state = 2;

            completer.complete(PreparedStmt._(
              preparedPacket: preparedPacket!,
              connection: this,
              iterable: iterable,
            ));

            _state = _MySQLConnectionState.connectionEstablished;

            return;
          }

          break;
      }

      if (packet != null) {
        final payload = packet.payload;

        if (payload is MySQLPacketStmtPrepareOK) {
          preparedPacket = payload;
        } else if (payload is MySQLPacketError) {
          throw Exception("MySQL error: ${payload.errorMessage}");
        } else {
          completer.completeError(
            "Unexpected payload received in response to COMM_STMT_PREPARE request",
          );
        }
      }
    };

    _socket.add(packet.encode());

    return completer.future;
  }

  Future<IResultSet> _executePreparedStmt(
    PreparedStmt stmt,
    List<dynamic> params,
    bool iterable,
  ) async {
    if (!_connected) {
      throw Exception("Can not execute prepared stmt: connecion closed");
    }

    // wait for ready state
    if (_state != _MySQLConnectionState.connectionEstablished) {
      await _waitForState(_MySQLConnectionState.connectionEstablished)
          .timeout(Duration(seconds: 10));
    }

    _state = _MySQLConnectionState.waitingCommandResponse;

    final payload = MySQLPacketCommStmtExecute(
      stmtID: stmt._preparedPacket.stmtID,
      params: params,
    );

    final packet = MySQLPacket(
      sequenceID: 0,
      payload: payload,
      payloadLength: 0,
    );

    final completer = Completer<IResultSet>();

    /**
     * 0 - initial
     * 1 - columnCount decoded
     * 2 - columnDefs parsed
     * 3 - eofParsed
     * 4 - rowsParsed
     */
    int state = 0;
    int colsCount = 0;
    List<MySQLColumnDefinitionPacket> colDefs = [];
    List<MySQLBinaryResultSetRowPacket> resultSetRows = [];

    // support for iterable result set
    IterablePreparedStmtResultSet? iterableResultSet;
    StreamSink<ResultSetRow>? sink;

    _responseCallback = (data) {
      MySQLPacket? packet;

      switch (state) {
        case 0:
          // if packet is OK packet, there is no data
          if (MySQLPacket.detectPacketType(data) == MySQLGenericPacketType.ok) {
            final okPacket = MySQLPacket.decodeGenericPacket(data);
            _state = _MySQLConnectionState.connectionEstablished;

            completer.complete(
              EmptyResultSet(okPacket: okPacket.payload as MySQLPacketOK),
            );

            return;
          }

          packet = MySQLPacket.decodeColumnCountPacket(data);
          break;
        case 1:
          packet = MySQLPacket.decodeColumnDefPacket(data);
          break;
        case 2:
          packet = MySQLPacket.decodeGenericPacket(data);
          if (packet.isEOFPacket()) {
            state = 3;
          } else if (packet.isErrorPacket()) {
            final errMsg = (packet.payload as MySQLPacketError).errorMessage;
            throw Exception("MySQL exception: $errMsg");
          } else {
            throw Exception("Unexcpected packet type");
          }
          break;
        case 3:
          if (iterable) {
            if (iterableResultSet == null) {
              iterableResultSet = IterablePreparedStmtResultSet._(
                columns: colDefs,
              );

              sink = iterableResultSet!._sink;
              completer.complete(iterableResultSet);
            }

            // check eof
            if (MySQLPacket.detectPacketType(data) ==
                MySQLGenericPacketType.eof) {
              state = 4;

              _state = _MySQLConnectionState.connectionEstablished;
              sink!.close();
              return;
            }

            packet = MySQLPacket.decodeBinaryResultSetRowPacket(data, colDefs);
            final values =
                (packet.payload as MySQLBinaryResultSetRowPacket).values;
            sink!.add(ResultSetRow._(colDefs: colDefs, values: values));
            packet = null;
            break;
          } else {
            // check eof
            if (MySQLPacket.detectPacketType(data) ==
                MySQLGenericPacketType.eof) {
              state = 4;

              final resultSetPacket = MySQLPacketBinaryResultSet(
                columnCount: BigInt.from(colsCount),
                columns: colDefs,
                rows: resultSetRows,
              );

              _state = _MySQLConnectionState.connectionEstablished;

              completer.complete(
                PreparedStmtResultSet._(resultSetPacket: resultSetPacket),
              );

              return;
            }

            packet = MySQLPacket.decodeBinaryResultSetRowPacket(data, colDefs);

            break;
          }
      }

      if (packet != null) {
        final payload = packet.payload;

        if (payload is MySQLPacketError) {
          completer.completeError("MySQL error: " + payload.errorMessage);
          return;
        } else if (payload is MySQLPacketOK) {
          // do nothing
        } else if (payload is MySQLPacketColumnCount) {
          state = 1;
          colsCount = payload.columnCount.toInt();
          return;
        } else if (payload is MySQLColumnDefinitionPacket) {
          colDefs.add(payload);
          if (colDefs.length == colsCount) {
            state = 2;
          }
        } else if (payload is MySQLBinaryResultSetRowPacket) {
          resultSetRows.add(payload);
        } else {
          completer.completeError(
            "Unexpected payload received in response to COMM_QUERY request",
          );
        }
      }
    };

    _socket.add(packet.encode());

    return completer.future;
  }

  Future<void> _deallocatePreparedStmt(PreparedStmt stmt) async {
    if (!_connected) {
      throw Exception("Can not execute query: connecion closed");
    }

    // wait for ready state
    if (_state != _MySQLConnectionState.connectionEstablished) {
      await _waitForState(_MySQLConnectionState.connectionEstablished)
          .timeout(Duration(seconds: 10));
    }

    final payload = MySQLPacketCommStmtClose(
      stmtID: stmt._preparedPacket.stmtID,
    );

    final packet = MySQLPacket(
      sequenceID: 0,
      payload: payload,
      payloadLength: 0,
    );

    _socket.add(packet.encode());
  }

  String _escapeString(String value) {
    value = value.replaceAll(r"\", r'\\');
    value = value.replaceAll(r"'", r"''");
    return value;
  }

  /// Close this connection
  ///
  /// This is an error to use this connection after connection has been closed
  Future<void> close() async {
    final packet = MySQLPacket(
      sequenceID: 0,
      payload: MySQLPacketCommQuit(),
      payloadLength: 0,
    );

    _socket.add(packet.encode());
    _state = _MySQLConnectionState.quitCommandSend;

    if (_socketSubscription != null) {
      await _socketSubscription!.cancel();
    }

    _connected = false;
    await _socket.flush();
    await Future.delayed(Duration(milliseconds: 10));
    await _socket.close();
    _socket.destroy();

    for (var element in _onCloseCallbacks) {
      element();
    }
    _onCloseCallbacks.clear();
  }

  Future<void> _waitForState(_MySQLConnectionState state) async {
    if (_state == state) {
      return;
    }

    await Future.doWhile(() async {
      if (_state == state) {
        return false;
      }

      await Future.delayed(Duration(microseconds: 100));
      return true;
    });
  }
}

/// Base class to represent result of calling [MySQLConnection.execute] and [PreparedStmt.execute]
abstract class IResultSet {
  /// Number of colums in this result if any
  int get numOfColumns;

  /// Number of rows in this result if any (unavailable for iterable results)
  int get numOfRows;

  /// Number of affected rows
  BigInt get affectedRows;

  /// Last insert ID
  BigInt get lastInsertID;

  /// Provides access to data rows (unavailable for iterable results)
  Iterable<ResultSetRow> get rows;

  /// Use [cols] to get info about returned columns
  Iterable<ResultSetColumn> get cols;

  /// Provides Stream like access to data rows. Use [rowsStream] to get rows from iterable results
  Stream<ResultSetRow> get rowsStream => Stream.fromIterable(rows);
}

/// Represents result of [MySQLConnection.execute] method
class ResultSet extends IResultSet {
  final MySQLPacketResultSet _resultSetPacket;

  ResultSet._({
    required MySQLPacketResultSet resultSetPacket,
  }) : _resultSetPacket = resultSetPacket;

  @override
  int get numOfColumns => _resultSetPacket.columns.length;

  @override
  int get numOfRows => _resultSetPacket.rows.length;

  @override
  BigInt get affectedRows => BigInt.zero;

  @override
  BigInt get lastInsertID => BigInt.zero;

  @override
  Iterable<ResultSetRow> get rows sync* {
    for (final _row in _resultSetPacket.rows) {
      yield ResultSetRow._(
        colDefs: _resultSetPacket.columns,
        values: _row.values,
      );
    }
  }

  @override
  Iterable<ResultSetColumn> get cols {
    return _resultSetPacket.columns.map(
      (e) => ResultSetColumn(
        name: e.name,
        type: e.type,
      ),
    );
  }
}

/// Represents result of [MySQLConnection.execute] method when passing iterable = true
class IterableResultSet implements IResultSet {
  final List<MySQLColumnDefinitionPacket> _columns;
  late StreamController<ResultSetRow> _controller;

  IterableResultSet._({
    required List<MySQLColumnDefinitionPacket> columns,
  }) : _columns = columns {
    _controller = StreamController();
  }

  StreamSink<ResultSetRow> get _sink => _controller.sink;

  @override
  Stream<ResultSetRow> get rowsStream => _controller.stream;

  @override
  int get numOfColumns => _columns.length;

  @override
  int get numOfRows => throw UnimplementedError(
        "numOfRows is not implemented for IterableResultSet",
      );

  @override
  BigInt get affectedRows => BigInt.zero;

  @override
  BigInt get lastInsertID => BigInt.zero;

  @override
  Iterable<ResultSetColumn> get cols {
    return _columns.map(
      (e) => ResultSetColumn(
        name: e.name,
        type: e.type,
      ),
    );
  }

  @override
  Iterable<ResultSetRow> get rows => throw UnimplementedError(
        "Use rowsStream to get rows from IterableResultSet",
      );
}

/// Represents result of [PreparedStmt.execute] method
class PreparedStmtResultSet extends IResultSet {
  final MySQLPacketBinaryResultSet _resultSetPacket;

  PreparedStmtResultSet._({
    required MySQLPacketBinaryResultSet resultSetPacket,
  }) : _resultSetPacket = resultSetPacket;

  @override
  int get numOfColumns => _resultSetPacket.columns.length;

  @override
  int get numOfRows => _resultSetPacket.rows.length;

  @override
  BigInt get affectedRows => BigInt.zero;

  @override
  BigInt get lastInsertID => BigInt.zero;

  @override
  Iterable<ResultSetRow> get rows sync* {
    for (final _row in _resultSetPacket.rows) {
      yield ResultSetRow._(
        colDefs: _resultSetPacket.columns,
        values: _row.values,
      );
    }
  }

  @override
  Iterable<ResultSetColumn> get cols {
    return _resultSetPacket.columns.map(
      (e) => ResultSetColumn(
        name: e.name,
        type: e.type,
      ),
    );
  }
}

/// Represents result of [PreparedStmt.execute] method when using iterable = true
class IterablePreparedStmtResultSet extends IResultSet {
  final List<MySQLColumnDefinitionPacket> _columns;
  late StreamController<ResultSetRow> _controller;

  IterablePreparedStmtResultSet._({
    required List<MySQLColumnDefinitionPacket> columns,
  }) : _columns = columns {
    _controller = StreamController();
  }

  StreamSink<ResultSetRow> get _sink => _controller.sink;

  @override
  int get numOfColumns => _columns.length;

  @override
  int get numOfRows => throw UnimplementedError(
        "numOfRows is not implemented for IterableResultSet",
      );

  @override
  BigInt get affectedRows => BigInt.zero;

  @override
  BigInt get lastInsertID => BigInt.zero;

  @override
  Iterable<ResultSetRow> get rows => throw UnimplementedError(
        "Use rowsStream to get rows from IterablePreparedStmtResultSet",
      );

  @override
  Stream<ResultSetRow> get rowsStream => _controller.stream;

  @override
  Iterable<ResultSetColumn> get cols {
    return _columns.map(
      (e) => ResultSetColumn(
        name: e.name,
        type: e.type,
      ),
    );
  }
}

/// Represents empty result set
class EmptyResultSet extends IResultSet {
  final MySQLPacketOK _okPacket;

  EmptyResultSet({required MySQLPacketOK okPacket}) : _okPacket = okPacket;

  @override
  int get numOfColumns => 0;

  @override
  int get numOfRows => 0;

  @override
  BigInt get affectedRows => _okPacket.affectedRows;

  @override
  BigInt get lastInsertID => _okPacket.lastInsertID;

  @override
  Iterable<ResultSetRow> get rows => List<ResultSetRow>.empty();

  @override
  Iterable<ResultSetColumn> get cols => List<ResultSetColumn>.empty();
}

/// Represents result set row data
class ResultSetRow {
  final List<MySQLColumnDefinitionPacket> _colDefs;
  final List<String?> _values;

  ResultSetRow._({
    required List<MySQLColumnDefinitionPacket> colDefs,
    required List<String?> values,
  })  : _colDefs = colDefs,
        _values = values;

  /// Get number of columns for this row
  int get numOfColumns => _colDefs.length;

  /// Get column data by column index (starting form 0)
  String colAt(int colIndex) {
    if (colIndex >= _values.length) {
      throw Exception("Column index is out of range");
    }

    final value = _values[colIndex]!;

    return value;
  }

  /// Get column data by column name
  String colByName(String columnName) {
    final colIndex =
        _colDefs.indexWhere((element) => element.name == columnName);

    if (colIndex == -1) {
      throw Exception("There is no column with name: $columnName");
    }

    if (colIndex >= _values.length) {
      throw Exception("Column index is out of range");
    }

    final value = _values[colIndex]!;

    return value;
  }

  /// Get data for all columns
  Map<String, String?> assoc() {
    final result = <String, String?>{};

    int colIndex = 0;

    for (final colDef in _colDefs) {
      result[colDef.name] = _values[colIndex];
      colIndex++;
    }

    return result;
  }
}

/// Represents column definition
class ResultSetColumn {
  String name;
  int type;

  ResultSetColumn({
    required this.name,
    required this.type,
  });
}

/// Prepared statement class
class PreparedStmt {
  final MySQLPacketStmtPrepareOK _preparedPacket;
  final MySQLConnection _connection;
  final bool _iterable;

  PreparedStmt._({
    required MySQLPacketStmtPrepareOK preparedPacket,
    required MySQLConnection connection,
    required bool iterable,
  })  : _preparedPacket = preparedPacket,
        _connection = connection,
        _iterable = iterable;

  int get numOfParams => _preparedPacket.numOfParams;

  /// Executes this prepared statement with given [params]
  Future<IResultSet> execute(List<dynamic> params) async {
    if (numOfParams != params.length) {
      throw Exception(
        "Can not execute prepared stmt: number of passed params != number of prepared params",
      );
    }

    return _connection._executePreparedStmt(this, params, _iterable);
  }

  /// Deallocates this prepared statement
  ///
  /// Use this method to prevent memory leaks for long running connections
  /// All prepared statements are automatically deallocated by database when connection is closed
  Future<void> deallocate() {
    return _connection._deallocatePreparedStmt(this);
  }
}
